import socket
from typing import Optional

import requests

from ovos_utils.log import LOG

_DEFAULT_TEST_CONFIG = {
    "ip_url": 'https://api.ipify.org',
    "dns_primary": "1.1.1.1",
    "dns_secondary": "8.8.8.8",
    "web_url": "http://nmcheck.gnome.org/check_network_status.txt",
    "web_url_secondary": "https://checkonline.home-assistant.io/online.txt",
    "captive_portal_url": "http://nmcheck.gnome.org/check_network_status.txt",
    "captive_portal_text": "NetworkManager is online"
}


def get_network_tests_config():
    """Get network_tests object from mycroft.configuration."""
    try:
        from ovos_config.config import read_mycroft_config
        config = read_mycroft_config()
    except ImportError:
        LOG.warning("ovos_config not available. Falling back to default config")
        config = dict()
    return config.get("network_tests", _DEFAULT_TEST_CONFIG)


def get_ip() -> str:
    """
    Get the local IPv4 address of this device
    taken from https://stackoverflow.com/a/28950776/13703283
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(10)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        ip = s.getsockname()[0]
    except Exception as e:
        LOG.error(e)
        ip = '127.0.0.1'
    finally:
        s.close()
    return ip


def get_external_ip() -> str:
    """
    Get the public IPv4 address of this device. If a public IP address cannot be
    determined, returns the public localhost address: `0.0.0.0`
    """
    cfg = get_network_tests_config()
    try:
        resp = requests.get(cfg.get("ip_url") or
                            _DEFAULT_TEST_CONFIG['ip_url'])
        if resp.ok:
            return resp.text
        LOG.error(f"Got resp={resp.status_code}: {resp.text}")
    except Exception as e:
        LOG.error(f"Unable to get external IP Address: {e}")
    return "0.0.0.0"


def is_connected_dns(host: Optional[str] = None, port: int = 53,
                     timeout: int = 3) -> bool:
    """
    Check internet connection by connecting to DNS servers
    Returns:
        True if internet connection can be detected
    """

    if host is None:
        cfg = get_network_tests_config()
        return is_connected_dns(cfg.get("dns_primary" or _DEFAULT_TEST_CONFIG['dns_primary'])) or \
            is_connected_dns(cfg.get("dns_secondary" or _DEFAULT_TEST_CONFIG['dns_secondary']))

    try:
        # connect to the host -- tells us if the host is actually reachable
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(timeout)
        s.connect((host, port))
        return True
    except OSError:
        pass
    return False


def is_connected_http(host: Optional[str] = None) -> bool:
    """
    Check internet connection by connecting to some website
    Returns:
        True if connection attempt succeeded
    """
    if host is None:
        cfg = get_network_tests_config()
        return is_connected_http(cfg.get("web_url") or
                                 _DEFAULT_TEST_CONFIG['web_url']) or \
            is_connected_http(cfg.get("web_url_secondary") or
                              _DEFAULT_TEST_CONFIG['web_url_secondary'])

    try:
        status = requests.head(host).status_code
        return True
    except:
        pass
    return False


def is_connected() -> bool:
    """
    Return True if any connection check (DNS or HTTP) is True
    """
    return any((is_connected_dns(), is_connected_http()))


def check_captive_portal(host: Optional[str] = None,
                         expected_text: Optional[str] = None) -> bool:
    """
    Returns True if a captive portal page is detected
    """
    captive_portal = False

    if not host or not expected_text:
        cfg = get_network_tests_config()
        host = host or cfg.get("captive_portal_url") or _DEFAULT_TEST_CONFIG["captive_portal_url"]
        expected_text = expected_text or cfg.get("captive_portal_text") or _DEFAULT_TEST_CONFIG["captive_portal_text"]

    try:
        # We need to check a site that doesn't use HTTPS
        html_doc = requests.get(host).text
        # If something different is in the html, we likely were redirected
        # to the portal page.
        if expected_text not in html_doc:
            captive_portal = True
    except Exception:
        LOG.exception("Error checking for captive portal")

    return captive_portal



import re
import os
from datetime import datetime
from traceback import FrameSummary
from dataclasses import dataclass
from typing import Any, Tuple, List, Generator, Dict, Union, Optional

from dateutil.parser import parse
import rich_click as click
from rich.console import Console
from rich.style import Style
from rich.table import Table
import pydoc
from combo_lock import ComboLock

try:
    from ovos_config import Configuration
    use24h = Configuration().get("time_format", "full") == "full"
    date_format = Configuration().get("date_format", "DMY")
except ImportError:
    use24h = True
    date_format = "DMY"
    
from ovos_utils.log import get_log_path, get_log_paths, get_available_logs


TIME_FORMAT = '%Y-%m-%d %H:%M:%S.%f'
LOGLOCK = ComboLock("ovos_logs_console_script")


@dataclass
class LogLine:
    timestamp: datetime = None
    source: str = ""
    location: str = ""
    level: str = ""
    message: str = ""

    def __str__(self):
        # sytsem messages etc.
        if not all([self.source, self.location, self.level]):
            return self.message
        return f"{self.format_timestamp()} - {self.source} - {self.location} - {self.level} - {self.message}"
    
    def format_timestamp(self):
        if self.timestamp:
            return self.timestamp.strftime(TIME_FORMAT)[:-3]
        return ""


# Traceback frame
class Frame(FrameSummary):
    def __init__(self, filename, lineno, name, line):
        super().__init__(filename, lineno, name, line=line)
      
    def as_dict(self):
        return {
            "location": self.format_location(),
            "level": "TRACEBACK",
            "message": self.line
        }
    
    def as_logline(self):
        return LogLine(**self.as_dict())
    
    def format_location(self):
        if "/bin/" in self.filename:
            package = self.filename.split("/bin/")[-1].replace(".py", "")\
                    .replace("-", "_").replace("/", ".")
        elif "site-packages" not in self.filename and \
                (pyver := re.search(r"python\d\.\d+[\\/]", self.filename)):
            package = self.filename.split(pyver.group())[-1].replace(".py", "")\
                    .replace("-", "_").replace("/", ".")
        else:
            package = self.filename.split("site-packages/")[-1].replace(".py", "")\
                    .replace("-", "_").replace("/", ".")
        method = self.name.replace(".py", "").replace("-", "_")
        return f"{package}:{method}:{self.lineno}"
    
    def __str__(self):
        return f'  File "{self.filename}", line {self.lineno}, in {self.name}\n    {self.line}\n'


class Traceback:
    PATTERN = r'File "(?P<filename>[^"]+)", line (?P<lineno>\d+), in (?P<name>\S+)\n\s*(?P<line>.+)'

    def __init__(self, frames: List[Frame], exception: str, timestamp: datetime = None):
        self.frames = frames
        self.exception = exception
        self._timestamp = timestamp
    
    @property
    def exception_location(self):
        return self.frames[-1].format_location()
    
    @property
    def timestamp(self):
        return self._timestamp
    
    @timestamp.setter
    def timestamp(self, value):
        self._timestamp = value

    def to_loglines(self) -> List[LogLine]:
        
        lines = [LogLine(timestamp=self.timestamp,
                         location=self.exception_location,
                         level="EXCEPTION",
                         message=self.exception)]

        for frame in self.frames:
            lines.append(frame.as_logline())
        
        return lines

    @classmethod
    def from_list(cls, lines):
        lines = [line if line.endswith("\n") else line + "\n" for line in lines]
        multiline = "".join(lines)
        return cls.from_string(multiline)
    
    @classmethod
    def from_string(cls, s):
        matches = re.findall(cls.PATTERN, s, re.MULTILINE)
        frames = []
        for match in matches:
            data = dict(zip(["filename", "lineno", "name", "line"], match))
            frames.append(Frame(**data))
        exception = next(line for line in s.split("\n")[::-1] if line)
        return cls(frames, exception)
    
    def __str__(self):
        multiline = "Traceback (most recent call last):\n"
        for frame in self.frames:
            multiline += str(frame)
        multiline += f"{self.exception}\n"
        return multiline


class OVOSLogParser:
    LOG_PATTERN = r'(?P<timestamp>\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{1,6}) - (?P<source>.+?) - (?P<location>.+?) - (?P<level>\w+) - (?P<message>.*)'

    @classmethod
    def parse(self, log_line, last_timestamp=None) -> LogLine:
        log_line.rstrip("\n")
        match = re.match(self.LOG_PATTERN, log_line)
        data = {}
        if match:
            data = match.groupdict()
            data['timestamp'] = datetime.strptime(data['timestamp'], TIME_FORMAT)
            return LogLine(**data)
        
        data["timestamp"] = last_timestamp or ""
        data["message"] = log_line
        return LogLine(**data)
    
    @classmethod
    def parse_file(self, source) -> Generator[Union[LogLine, Traceback], None, None]:
        if not os.path.exists(source):
            raise FileNotFoundError(f"File {source} does not exist")

        with open(source, 'r') as file:
            trace = None
            last_timestamp = None
            for line in file:
                # gather all lines of the traceback
                if line == "Traceback (most recent call last):\n":
                    trace = [line]
                    continue
                # TODO do tracebacks always end on a empty line?
                elif trace and line == "\n":
                    trace.append(line)
                    traceback = Traceback.from_list(trace)
                    traceback.timestamp = last_timestamp
                    yield traceback
                    trace = None
                elif trace:
                    trace.append(line)
                else:
                    log = self.parse(line, last_timestamp)
                    if log.message == "\n":
                        continue
                    timestamp = log.timestamp
                    if timestamp:
                        last_timestamp = timestamp
                    yield log


console = Console()

EXPECTED_DATE_FORMAT = "YYYY-MM-DD" if date_format == "YMD" else "DD-MM-YYYY"
EXPECTED_DATE = "2023-12-01" if date_format == "YMD" else "01-12-2023"
EXPECTED_DATETIME_FORMAT = f'"[{EXPECTED_DATE_FORMAT}] HH:MM[:SS]{" AM/PM" if not use24h else ""}"'
EXPECTED_TIME = f'"09:00:05{" PM" if not use24h else ""}"'

LOGSOPTHELP = """logs to be sliced 
\n\nmultiple: -l bus -l audio"""
STARTTIMEHELP = f"""start time of the log slice (default: since service restart,
input format: {EXPECTED_DATETIME_FORMAT})
\n\nExample: -s \"{EXPECTED_DATE} 12:00{' AM/PM' if not use24h else ''}\" / -s
 {'"' if not use24h else ''}12:00:05{' AM/PM"' if not use24h else ''}"""

click.rich_click.STYLE_ARGUMENT = "dark_red"
click.rich_click.STYLE_OPTION = "dark_red"
click.rich_click.STYLE_SWITCH = "indian_red"
click.rich_click.USE_MARKDOWN = True
click.rich_click.COMMAND_GROUPS = {
    "ovos-logs": [
        {
            "name": "Slice logs by time",
            "commands": ["slice"],
            "table_styles": {
                "row_styles": ["white"],
                "padding": (0, 2),
                "title_justify": "left"
            },
        },
        {
            "name": "List logs by severity",
            "commands": ["list"],
            "table_styles": {
                "row_styles": ["white"],
                "padding": (0, 2),
            },
        },
        {
            "name": "Downsize logs",
            "commands": ["reduce"],
            "table_styles": {
                "row_styles": ["white"],
                "padding": (0, 1),
            },
        },
        {
            "name": "Show logs (using less)",
            "commands": ["show"],
            "table_styles": {
                "row_styles": ["white"],
                "padding": (0, 2),
            },
        }
    ]
}


def get_last_load_time(directories: Optional[List[str]] = None) -> Optional[datetime]:
    # if nothing's found return the beginning of unix time
    last_timestamp = datetime.fromtimestamp(0)
    if directories is None:
        directory = get_log_path("skills")
    else:
        directory = get_log_path("skills", directories)
    
    if directory:
        with open(os.path.join(directory,"skills.log"), "r") as f:
            for line in f.readlines()[::-1]:
                logline = OVOSLogParser.parse(line)
                if logline.timestamp:
                    last_timestamp = logline.timestamp 
                if logline.message == "Loading message bus configs":
                    break
    return last_timestamp


def valid_log(logs, paths):
    for log in logs:
        if log.lower() not in get_available_logs(paths):
            return False
    return True


def parse_time(time_str):
    try:
        time = parse(time_str)
    except ValueError:
        return None
    return time


def get_timestamped_filename(basename: str, ext: str, basedir = "~", timeformat = '%Y%m%d_%H%M%S'):
    if basedir == "~":
        basedir = os.path.expanduser("~")

    t = datetime.now().strftime(timeformat)
    return os.path.join(basedir, f"{basename}_{t}.{ext}")


def parse_timeframe(start, end, directories: Optional[List[str]] = None) -> Tuple[Any, Any]:
    """
    Parses the start and end time given a string input.
    If the start is None, parse the skill log to determine the last service load time and
    if that fails return the beginning starting datetime of the log.
    If the end is None, return the current datetime. 

    :param start: start time of the log slice (default: since service restart)
    :param end: end time of the log slice (default: now)
    :param directories: the directory logs reside in
    :return: start and end time
    """
    if start is None:
        start = get_last_load_time(directories)
    else:
        start = parse_time(start)
    
    if end is None:
        end = datetime.now()
    else:
        end = parse_time(end)
    return start, end


@click.group()
def ovos_logs():
    """\b
    Small helper tool to quickly navigate the logs, create slices and quickview errors

    `ovos-logs [COMMAND] --help` for further information about the specific command ARGUMENTS
    \b
    """
    pass


@ovos_logs.command()
@click.option("--start", "-s", help=STARTTIMEHELP)
@click.option("--until", "-u", help=f"end time of the log slice [default: now]")
@click.option("--logs", "-l", multiple=True, default=get_available_logs(), help=LOGSOPTHELP, show_default=True)
@click.option("--paths", "-p", multiple=True, default=get_log_paths(), help=f"the directory logs reside in", show_default=True)
@click.option("--file", "-f", is_flag=False, flag_value=get_timestamped_filename("slice", "log"),
              default=None, help=f"output as file (if flagged, but not specified: {get_timestamped_filename('slice', 'log')})")
def slice(start, until, logs, paths, file):
    """\b
    Optionally define start (`-s`) and the time until (`-u`) the slice should be limited to.  
    \b
    Different logs can be included using the `-l` option. If not specified, all logs will be included.  
    Optionally the directory where the logs are stored (`-p`) and the file where the slices should be dumped (`-f`)
    can be specified.  
    \b
    > Examples:  
    > ovos-logs slice                                            # Slice all logs from service start up until now  
    > ovos-logs slice -s 01-12-2023 -u '01-12-2023 17:00:20'     # Slice all logs from the start of december the first until 17:00:20  
    > ovos-logs slice -l bus -l skills -f ~/myslice.log          # Slice skills.log and bus.log from service start up until now and dump it to the file ~/myslice.log  
    """
    logs_present = []

    if not all(os.path.exists(path) for path in paths):
        return console.print(f"Directory [{[p for p in paths if not os.path.exists(p)]}] does not exist")
    else:
        logs_present = get_available_logs(paths)

    start, end = parse_timeframe(start, until, paths)
    if start is None:
        return console.print(f"Need a valid start time in the format {EXPECTED_DATETIME_FORMAT}")
    elif end is None:
        return console.print(f"Need a valid end time in the format {EXPECTED_DATETIME_FORMAT}")
    elif start > end:
        return console.print(f"Start time [{start}] is after end time [{end}]")

    if not logs:
        logs = logs_present
    elif not valid_log(logs, paths):
        return console.print(f"Invalid log name, valid logs are {logs_present}")

    _templog: Dict[str, List[LogLine]] = dict()

    for service in logs:
        path = get_log_path(service, paths)
        logfile = os.path.join(path, f"{service}.log")
        if not os.path.exists(logfile):
            continue
        _templog[service] = []
        for log in OVOSLogParser.parse_file(logfile):
            if start <= log.timestamp < end:
                if isinstance(log, Traceback):
                    _templog[service].extend(log.to_loglines())
                else:
                    _templog[service].append(log)
        if not _templog[service]:
            del _templog[service]

    if not _templog:
        return console.print("No logs found in the specified time frame")
    
    if file is not None:
        # test if file is writable
        try:
            with open(file, 'w') as f:
                pass
        except:
            return console.print(f"File [{file}] is not writable. Aborted")
        else:
            console.print(f"Log slice saved to [bold]{file}[/bold]")

    for service in _templog:
        table = Table(title=service)
        table.add_column("Time", style="cyan", no_wrap=True)
        table.add_column()
        table.add_column("Message", style="magenta")
        table.add_column("Origin", style="green")
        lineno = 0
        for logline in _templog[service]:
            lineno += 1
            style = None
            timestamp = logline.timestamp or ""
            if isinstance(timestamp, datetime):
                timestamp = timestamp.strftime("%H:%M:%S.%f" if use24h else "%I:%M:%S.%f")[:-3]
                if not use24h:
                    timestamp += logline.timestamp.strftime(" %p")

            level = logline.level or ""
            message = logline.message or ""
            if level == "ERROR":
                level = "[bold red]" + level[:1]
            elif level == "EXCEPTION":
                level = "[bold red]" + level[:3]
            elif level == "WARNING":
                level = "[bold yellow]" + level[:1]
            elif level == "DEBUG":
                level = "[bold blue]" + level[:1]
            elif level == "TRACEBACK":
                level = "[white]" + level[:5]
                message = "[grey42]" + message
            elif level == "INFO":
                level = ""
                message = "[navajo_white1]" + message
            if lineno % 2 == 0:
                style = Style(bgcolor="grey7")
            table.add_row(
                timestamp,
                level,
                message,
                logline.location or "",
                style=style
            )
            if len(logline.message) > 200:
                table.add_row()

        console.print(table)
        if file:
            Console(file=open(file, 'a')).print(table)


@ovos_logs.command()
@click.option("--error", "-e", is_flag=True, help="display error messages")
@click.option("--warning", "-w", is_flag=True, help="display warning messages")
@click.option("--exception", "-x", is_flag=True, help="display exceptions")
@click.option("--debug", "-d", is_flag=True, help="display debug messages")
@click.option("--start", "-s", help=STARTTIMEHELP)
@click.option("--until", "-u", help=f"end time of the log slice [default: now]")
@click.option("--logs", "-l", multiple=True, default=get_available_logs(), help=LOGSOPTHELP, show_default=True)
@click.option("--paths", "-p", multiple=True, type=click.Path(), default=get_log_paths(), help=f"the directory logs reside in", show_default=True)
@click.option("--file", "-f", is_flag=False, type=click.Path(), flag_value=get_timestamped_filename("list", "log"), default=None,
              help=f"output as file (if flagged, but not specified: {get_timestamped_filename('list', 'log')})")
def list(error, warning, exception, debug, start, until, logs, paths, file):
    """\b
    Log level has to be specified.  
    \b
    Optionally define start (`-s`) and the time until (`-u`) the slice should be limited to.  
    \b
    Different logs can be included using the `-l` option. If not specified, all logs will be included.  
    \b
    Optionally the directory where the logs are stored (`-p`) and the file where the slices should be dumped (`-f`)
    can be specified.  
    \b
    > Examples:  
    > ovos-logs list -x                                           # List all exceptions from service start up until now  
    > ovos-logs list -e -w -s 01-12-2023 -u '01-12-2023 17:00:20' # List all errors and warnings from the start of december the first until 17:00:20  
    > ovos-logs list -x -l bus -l skills -f                       # List all exceptions from skills.log and bus.log and dump it to the file ~/list_xxx_xxx.log  
    """
    if not any([error, warning, debug, exception]):
        return console.print("Need at least one of --error, --warning, --exception or --debug")
    else:
        log_levels = [lv_str for lv, lv_str in [(error, "ERROR"), (warning, "WARNING"),
                                                (debug, "DEBUG"), (exception, "EXCEPTION")] if lv]

    if not all(os.path.exists(path) for path in paths):
        return console.print(f"Directory [{[p for p in paths if not os.path.exists(p)]}] does not exist")
    else:
        logs_present = get_available_logs(paths)

    start, end = parse_timeframe(start, until, paths)
    if start is None:
        return console.print(f"Need a valid start time in the format {EXPECTED_DATETIME_FORMAT}")
    elif end is None:
        return console.print(f"Need a valid end time in the format {EXPECTED_DATETIME_FORMAT}")
    elif start > end:
        return console.print(f"Start time [{start}] is after end time [{end}]")
    
    if not logs:
        logs = logs_present
    elif not valid_log(logs, paths):
        return console.print(f"Invalid log name, valid logs are {logs_present}")
    
    _templog: Dict[str, List[LogLine]] = dict()

    for service in logs:
        path = get_log_path(service, paths)
        if path is None:
            continue
        logfile = os.path.join(path, f"{service}.log")
        _templog[service] = []
        for log in OVOSLogParser.parse_file(logfile):
            if isinstance(log, Traceback):
                if exception:
                    _templog[service].extend(log.to_loglines())
                continue
            # LOG.exception
            if exception and log.level == "EXCEPTION":
                _templog[service].append(log)
            if error and log.level == "ERROR":
                _templog[service].append(log)
            if warning and log.level == "WARNING":
                _templog[service].append(log)
            if debug and log.level == "DEBUG":
                _templog[service].append(log)
        if not _templog[service]:
            del _templog[service]
    
    if not _templog:
        return console.print("No logs found for the specified log level")
    
    if file is not None:
        # test if file is writable
        try:
            with open(file, 'w') as f:
                pass
        except:
            return console.print(f"File [{file}] is not writable. Aborted")
        else:
            console.print(f"Log list saved to [bold]{file}[/bold]")

    for service in _templog:
        table = Table(title=f"{service} ({','.join(log_levels)})")
        # for traceback indication
        table.add_column("Time", style="cyan", no_wrap=True)
        if exception or len(log_levels) > 1:
            table.add_column()
        table.add_column("Message", style="magenta")
        table.add_column("Origin", style="green")
        lineno = 0
        for log in _templog[service]:
            style = None
            lineno += 1
            timestamp = log.timestamp or ""
            if timestamp:
                timestamp = timestamp.strftime("%H:%M:%S.%f" if use24h else "%I:%M:%S.%f")[:-3]
            if not use24h and timestamp:
                timestamp += log.timestamp.strftime(" %p")
            level = log.level.upper()
            message = log.message.rstrip("\n")
            if level == "ERROR":
                level = "[bold red]" + level[:1]
            elif level == "EXCEPTION":
                level = "[bold red]" + level[:3]
            elif level == "WARNING":
                level = "[bold yellow]" + level[:1]
            elif level == "DEBUG":
                level = "[bold blue]" + level[:1]
            elif level == "TRACEBACK":
                level = "[white]" + level[:5]
                message = "[grey42]" + message
            elif level == "INFO":
                level = ""
                message = "[navajo_white1]" + message

            if lineno % 2 == 0:
                style = Style(bgcolor="grey7")
            row = [timestamp, level, message, log.location]
            if not exception and len(log_levels) < 2:
                row.pop(1)                
            table.add_row(*row, style=style)
            if len(log.message) > 200:
                table.add_row()

        console.print(table)
        if file:
            Console(file=open(file, 'a')).print(table)

    
@ovos_logs.command()
@click.option("--log", "-l", required=True, type=click.Choice(get_available_logs(), case_sensitive=False), help=f"log to show; available: {get_available_logs()}")
@click.option("--paths", "-p", multiple=True, type=click.Path(), default=get_log_paths(), help=f"the directory logs reside in", show_default=True)
def show(log, paths):
    """\b
    A service log has to be specified (`-l`).  
    \b
    Optionally the directory where the logs are stored (`-p`) can be specified.  
    \b
    > Examples:  
    > ovos-logs show -l skills                                    # Display skills.log   
    > ovos-logs show -l debug -p ~/custom_path/                   # Display debug.log from a custom path     
    """
    if not any(os.path.exists(os.path.join(path, f"{log}.log")) for path in paths):
        return console.print(f"File does not exist")
    else:
        log = os.path.join(get_log_path(log, paths), f"{log}.log")
    
    pydoc.pager(open(log).read())


@ovos_logs.command()
@click.option("--size", "-s", is_flag=False, flag_value=None, default=0, help="truncate logs to a given size (in bytes)")
@click.option("--date", "-d", help="truncate logs to a given date")
@click.option("--logs", "-l", multiple=True, default=get_available_logs(), help=LOGSOPTHELP, show_default=True)
@click.option("--paths", "-p", multiple=True, type=click.Path(), default=get_log_paths(), help=f"the directory logs reside in", show_default=True)
def reduce(size, date, logs, paths):
    """\b
    Reduce logs to a given size (in bytes) or remove entries before a given date.  
    \b
    Different logs can be included using the `-l` option. If not specified, all logs will be included.  
    Optionally the directory where the logs are stored (`-p`) can be specified.  
    \b
    > Examples:  
    > ovos-logs reduce                                            # Reduce all logs to 0 bytes  
    > ovos-logs reduce -s 1000000                                 # Reduce all logs to ~1MB (latest logs)  
    > ovos-logs reduce -d "1-12-2023 17:00"                       # Reduce all logs to entries after the specified date/time  
    > ovos-logs reduce -s 1000000 -l skills -l bus                # Reduce skills.log and bus.log to ~1MB (latest logs)  
    """

    if date:
        size = None
        date = parse_time(date)
        if date is None:
            return console.print(f"The date/time provided couldn't be parsed. Expected format: {EXPECTED_DATETIME_FORMAT}")
        
    if not all(os.path.exists(path) for path in paths):
        return console.print(f"Directory [{[p for p in paths if not os.path.exists(p)]}] does not exist")
    else:
        logs_present = get_available_logs(paths)
    
    if not logs:
        logs = logs_present
    elif not valid_log(logs, paths):
        return console.print(f"Invalid log name, valid logs are {logs_present}")
    
    for service in logs:
        path = get_log_path(service, paths)
        logfile = os.path.join(path, f"{service}.log")
        reduced = False
        with LOGLOCK:
            if size:
                with open(logfile, 'r') as f:
                    f.seek(0, os.SEEK_END)
                    fullsize = f.tell()
                    f.seek(max(fullsize - size, 0))
                    # skip cutoff line
                    f.readline()
                    remaining_lines = f.readlines()
                if fullsize > size and remaining_lines:
                    reduced = True
                    with open(logfile, 'w') as f:
                        f.writelines(remaining_lines)
            elif date:
                loglines = []
                for log in OVOSLogParser.parse_file(logfile):
                    if log.timestamp and log.timestamp < date:
                        reduced = True
                        continue
                    loglines.append(log)
                if reduced:
                    with open(logfile, 'w') as f:
                        for log in loglines:
                            f.write(str(log) + "\n")
            else:
                reduced = True
                with open(logfile, 'w') as f:
                    f.write("")

        if reduced:
            console.print(f"{service} log reduced")



# Mapping based on Jeffers phoneme to viseme map, seen in table 1 from:
# http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.221.6377&rep=rep1&type=pdf
#
# Mycroft unit visemes based on images found at:
# http://www.web3.lu/wp-content/uploads/2014/09/visemes.jpg
#
# Mapping was created partially based on the "12 mouth shapes visuals seen at:
# https://wolfpaulus.com/journal/software/lipsynchronization/

VISIMES = {
    # /A group
    'v': '5',
    'f': '5',
    # /B group
    'uh': '2',
    'w': '2',
    'uw': '2',
    'er': '2',
    'r': '2',
    'ow': '2',
    # /C group
    'b': '4',
    'p': '4',
    'm': '4',
    # /D group
    'aw': '1',
    # /E group
    'th': '3',
    'dh': '3',
    # /F group
    'zh': '3',
    'ch': '3',
    'sh': '3',
    'jh': '3',
    # /G group
    'oy': '6',
    'ao': '6',
    # /Hgroup
    'z': '3',
    's': '3',
    # /I group
    'ae': '0',
    'eh': '0',
    'ey': '0',
    'ah': '0',
    'ih': '0',
    'y': '0',
    'iy': '0',
    'aa': '0',
    'ay': '0',
    'ax': '0',
    'hh': '0',
    # /J group
    'n': '3',
    't': '3',
    'd': '3',
    'l': '3',
    # /K group
    'g': '3',
    'ng': '3',
    'k': '3',
    # blank mouth
    'pau': '4',
}



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import functools
import inspect
import json
import logging
import os
import sys
from logging.handlers import RotatingFileHandler
from os.path import join
from pathlib import Path
from typing import Optional, List, Set

ALL_SERVICES = {"bus",
                "audio",
                "skills",
                "voice",
                "gui",
                "ovos",
                "phal",
                "phal-admin",
                "hivemind",
                "hivemind-voice-sat"}


class LOG:
    """
    Custom logger class that acts like logging.Logger

    The logger name is gnerally set by the calling module, but the default name
    is read from the envvar `OVOS_DEFAULT_LOG_NAME`.

    The log level defaults to `INFO` and can be overridden by
    `OVOS_DEFAULT_LOG_LEVEL`. Note that log level may be overridden by
     configuration when calling `LOG.init`.

    The config file can have a "logging" section

    {
        "logging": {
	    "log_level": "INFO",  // default log level
            "logs": {
                "path": "/opt/ovos/logs/",
                "max_bytes": 50000000,
                "backup_count": 6
            }.
            "bus": {  // override for different services
	        "log_level": "DEBUG"
                "logs": { // optionally override default logs
                          // (nb this does not merge values so
                          // backup_count takes the default 3 here)
                      "path": "/path/for/just/bus/logs"
                }
	     },
        },
        ....
    }


    Usage:
        >>> LOG.debug('My message: %s', debug_str)
        13:12:43.673 - :<module>:1 - DEBUG - My message: hi
        >>> LOG('custom_name').debug('Another message')
        13:13:10.462 - custom_name - DEBUG - Another message
    """
    base_path = "stdout"
    fmt = '%(asctime)s.%(msecs)03d - ' \
          '%(name)s - %(levelname)s - %(message)s'
    datefmt = '%Y-%m-%d %H:%M:%S'
    formatter = logging.Formatter(fmt, datefmt)
    max_bytes = 50000000
    backup_count = 3
    name = os.getenv("OVOS_DEFAULT_LOG_NAME") or 'OVOS'
    level = os.getenv("OVOS_DEFAULT_LOG_LEVEL") or "INFO"
    diagnostic_mode = False
    _loggers = {}

    @classmethod
    def __init__(cls, name='OVOS'):
        cls.name = name

    @classmethod
    def init(cls, config=None):
        from ovos_utils.xdg_utils import xdg_state_home
        try:
            from ovos_config.meta import get_xdg_base
            xdg_base = get_xdg_base()
        except ImportError:
            xdg_base = os.environ.get("OVOS_CONFIG_BASE_FOLDER") or "mycroft"

        xdg_path = os.path.join(xdg_state_home(), xdg_base)

        config = config or {}
        cls.base_path = config.get("path") or xdg_path
        cls.max_bytes = config.get("max_bytes", 50000000)
        cls.backup_count = config.get("backup_count", 3)
        level = config.get("level") or LOG.level
        cls.set_level(level)
        cls.diagnostic_mode = config.get("diagnostic", False)

    @classmethod
    def create_logger(cls, name, tostdout=True):
        if name in cls._loggers:
            return cls._loggers[name]
        logger = logging.getLogger(name)
        logger.propagate = False
        # also log to stdout
        if tostdout or cls.base_path == "stdout":
            stdout_handler = logging.StreamHandler(sys.stdout)
            stdout_handler.setFormatter(cls.formatter)
            logger.addHandler(stdout_handler)
        # log to file
        if cls.base_path != "stdout":
            os.makedirs(cls.base_path, exist_ok=True)
            path = join(cls.base_path,
                        cls.name.lower().strip() + ".log")
            handler = RotatingFileHandler(path, maxBytes=cls.max_bytes,
                                          backupCount=cls.backup_count)
            handler.setFormatter(cls.formatter)
            logger.addHandler(handler)
        logger.setLevel(cls.level)
        cls._loggers[name] = logger
        return logger

    @classmethod
    def set_level(cls, level):
        cls.level = level
        for l in cls._loggers:
            cls._loggers[l].setLevel(level)

    @classmethod
    def _get_real_logger(cls):
        name = ""
        if cls.name is not None:
            name = cls.name + " - "

        # Stack:
        # [0] - _log()
        # [1] - debug(), info(), warning(), or error()
        # [2] - caller
        stack = inspect.stack()

        # Record:
        # [0] - frame object
        # [1] - filename
        # [2] - line number
        # [3] - function
        # ...
        record = stack[2]
        mod = inspect.getmodule(record[0])
        module_name = mod.__name__ if mod else ''
        name += module_name + ':' + record[3] + ':' + str(record[2])

        logger = cls.create_logger(name, tostdout=True)
        if cls.diagnostic_mode:
            try:
                from ovos_bus_client.message import dig_for_message
                msg = dig_for_message()
                if msg:
                    logger.debug(f"DIAGNOSTIC - source bus message {msg.serialize()}")
            except ImportError:
                pass
        return logger

    @classmethod
    def info(cls, *args, **kwargs):
        cls._get_real_logger().info(*args, **kwargs)

    @classmethod
    def debug(cls, *args, **kwargs):
        cls._get_real_logger().debug(*args, **kwargs)

    @classmethod
    def warning(cls, *args, **kwargs):
        cls._get_real_logger().warning(*args, **kwargs)

    @classmethod
    def error(cls, *args, **kwargs):
        cls._get_real_logger().error(*args, **kwargs)

    @classmethod
    def exception(cls, *args, **kwargs):
        cls._get_real_logger().exception(*args, **kwargs)


def _monitor_log_level():
    _logs_conf = get_logs_config(LOG.name)
    hax = hash(json.dumps(_logs_conf, sort_keys=True, indent=2))
    if hax != _monitor_log_level.config_hash:
        _monitor_log_level.config_hash = hax
        LOG.init(_logs_conf)
        LOG.info("updated LOG level")


_monitor_log_level.config_hash = None


def init_service_logger(service_name):
    _logs_conf = get_logs_config(service_name)
    _monitor_log_level.config_hash = hash(json.dumps(_logs_conf, sort_keys=True, indent=2))
    LOG.name = service_name
    LOG.init(_logs_conf)  # setup the LOG instance
    try:
        from ovos_config import Configuration
        Configuration.set_config_watcher(_monitor_log_level)
    except ImportError:
        LOG.warning("Can not monitor config LOG level changes")


def get_logs_config(service_name=None, _cfg=None) -> dict:
    if _cfg is None:
        try:
            from ovos_config import Configuration
            _cfg = Configuration()
        except ImportError:
            LOG.warning("ovos_config not available. Falling back to defaults")
            _cfg = {}

    # First try and get the "logging" section
    log_config = _cfg.get("logging")
    # For compatibility we try to get the "logs" from the root level
    # and default to empty which is used in case there is no logging
    # section
    _logs_conf = _cfg.get("logs") or {}
    if log_config:  # We found a logging section
        # if "logs" is defined in "logging" use that as the default
        # where per-service "logs" are not defined
        _logs_conf = log_config.get("logs") or _logs_conf
        # Now get our config by service name
        if service_name:
            _cfg = log_config.get(service_name) or log_config
            # and if "logs" is redefined in "logging.<service_name>" use that
            _logs_conf = _cfg.get("logs") or _logs_conf
    # Grab the log level from whatever section we found, defaulting to INFO
    _log_level = _cfg.get("log_level", "INFO")
    _logs_conf["level"] = _log_level
    return _logs_conf


def log_deprecation(log_message: str = "DEPRECATED",
                    deprecation_version: str = "Unknown",
                    func_name: str = None,
                    func_module: str = None,
                    excluded_package_refs: List[str] = None):
    """
    Log a deprecation warning with information for the call outside the module
    that is generating the warning
    @param log_message: Log contents describing the deprecation
    @param deprecation_version: package version in which method will be deprecated
    @param func_name: decorated function name (else read from stack)
    @param func_module: decorated function module (else read from stack)
    @param excluded_package_refs: list of packages to exclude from call origin
        determination. i.e. an internal exception handling method should log the
        first call external to that package
    """
    stack = inspect.stack()[1:]  # [0] is this method
    call_info = "Unknown Origin"
    origin_module = func_module
    log_name = f"{LOG.name} - {func_module}:{func_name}" if \
        func_module and func_name else LOG.name
    for call in stack:
        module = inspect.getmodule(call.frame)
        name = module.__name__ if module else call.filename
        if any((name if name.startswith(x) else None
                for x in ("ovos_utils.log", "<"))):
            # Skip calls from this module and unittests to get at real origin
            continue
        if not origin_module:
            # Assume first outside call is the origin if not specified
            origin_module = name
            log_name = f"{LOG.name} - {name}:{func_name or call[3]}:{call[2]}"
            continue
        if excluded_package_refs and any((name.startswith(x) for x in
                                          excluded_package_refs)):
            continue
        if not name.startswith(origin_module):
            call_info = f"{name}:{call.lineno}"
            break
    # Explicitly format log to print origin log reference
    LOG.create_logger(log_name).warning(
        f"Deprecation version={deprecation_version}. Caller={call_info}. "
        f"{log_message}")


def deprecated(log_message: str, deprecation_version: str):
    """
    Decorator to log deprecation on call to deprecated function
    @param log_message: Deprecation log message
    @param deprecation_version: package version in which deprecation will occur
    """

    def wrapped(func):
        @functools.wraps(func)
        def log_wrapper(*args, **kwargs):
            log_deprecation(log_message=log_message,
                            func_name=func.__qualname__,
                            func_module=func.__module__,
                            deprecation_version=deprecation_version)
            return func(*args, **kwargs)

        return log_wrapper

    return wrapped


def get_log_path(service: str, directories: Optional[List[str]] = None) \
        -> Optional[str]:
    """
    Get the path to the log directory for a given service.
    Default behaviour is to check the configured paths for the service.
    If a list of directories is provided, check that list for the service log

    Args:
        service: service name
        directories: (optional) list of directories to check for service

    Returns:
        path to log directory for service
    """
    if directories:
        for directory in directories:
            file = os.path.join(directory, f"{service}.log")
            if os.path.exists(file):
                return directory
        return None

    from ovos_utils.xdg_utils import xdg_state_home
    try:
        from ovos_config import Configuration
        from ovos_config.meta import get_xdg_base
    except ImportError:
        xdg_base = os.environ.get("OVOS_CONFIG_BASE_FOLDER", "mycroft")
        return os.path.join(xdg_state_home(), xdg_base)

    config = Configuration().get("logging", dict()).get("logs", dict())
    # service specific config or default config location
    path = config.get(service, {}).get("path") or config.get("path")
    # default xdg location
    if not path:
        path = os.path.join(xdg_state_home(), get_xdg_base())

    return path


def get_log_paths() -> Set[str]:
    """
    Get all log paths for all service logs
    Different services may have different log paths

    Returns:
        set of paths to log directories
    """
    paths = set()
    ALL_SERVICES.union({s.replace("-", "_") for s in ALL_SERVICES})
    for service in ALL_SERVICES:
        paths.add(get_log_path(service))

    return paths


def get_available_logs(directories: Optional[List[str]] = None) -> List[str]:
    """
    Get a list of all available log files
    Args:
        directories: (optional) list of directories to check for service

    Returns:
        list of log files
    """
    directories = directories or get_log_paths()
    return [Path(f).stem for path in directories
            for f in os.listdir(path) if Path(f).suffix == ".log"]



from collections import defaultdict
from xml.etree import cElementTree as ET


def etree2dict(t):
    d = {t.tag: {} if t.attrib else None}
    children = list(t)
    if children:
        dd = defaultdict(list)
        for dc in map(etree2dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if t.attrib:
        d[t.tag].update((k, v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]['text'] = text
        else:
            d[t.tag] = text
    return d


def xml2dict(xml_string):
    try:
        xml_string = xml_string.replace('xmlns="http://www.w3.org/1999/xhtml"',
                                        "")
        e = ET.XML(xml_string)
        d = etree2dict(e)
        return d
    except:
        return {}


def load_xml2dict(xml_path):
    tree = ET.parse(xml_path)
    d = etree2dict(tree.getroot())
    return d["xml"]


def dict2xml(d, root="xml"):
    xml = "<" + root
    # props
    for k in d:
        if isinstance(d[k], str):
            if k == "text":
                pass
            else:
                xml += " " + k + ' ="' + d[k] + '"'

    xml += ">"
    # insides
    for k in d:
        if isinstance(d[k], dict):
            xml += dict2xml(d[k], k)
        if isinstance(d[k], list):
            for e in d[k]:
                if isinstance(e, dict):
                    xml += dict2xml(e, k)
        if isinstance(d[k], str):
            if k == "text":
                xml += d[k]

    xml += "</" + root + ">"
    return xml



import os
from setuptools import setup

BASEDIR = os.path.abspath(os.path.dirname(__file__))


def get_version():
    """ Find the version of the package"""
    version = None
    version_file = os.path.join(BASEDIR, 'ovos_utils', 'version.py')
    major, minor, build, alpha, post = (None, None, None, None, None)
    with open(version_file) as f:
        for line in f:
            if 'VERSION_MAJOR' in line:
                major = line.split('=')[1].strip()
            elif 'VERSION_MINOR' in line:
                minor = line.split('=')[1].strip()
            elif 'VERSION_BUILD' in line:
                build = line.split('=')[1].strip()
            elif 'VERSION_ALPHA' in line:
                alpha = line.split('=')[1].strip()
            elif 'VERSION_POST' in line:
                post = line.split('=')[1].strip()

            if ((major and minor and build and alpha) or
                    '# END_VERSION_BLOCK' in line):
                break
    version = f"{major}.{minor}.{build}"
    if alpha and int(alpha) > 0:
        version += f"a{alpha}"
    elif post and int(post) > 0:
        version += f"post{post}"
    return version


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


def required(requirements_file):
    """ Read requirements file and remove comments and empty lines. """
    with open(os.path.join(BASEDIR, requirements_file), 'r') as f:
        requirements = f.read().splitlines()
        if 'MYCROFT_LOOSE_REQUIREMENTS' in os.environ:
            print('USING LOOSE REQUIREMENTS!')
            requirements = [r.replace('==', '>=').replace('~=', '>=') for r in requirements]
        return [pkg for pkg in requirements
                if pkg.strip() and not pkg.startswith("#")]

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='ovos_utils',
    version=get_version(),
    packages=['ovos_utils',
              'ovos_utils.lang'],
    url='https://github.com/OpenVoiceOS/ovos_utils',
    install_requires=required("requirements/requirements.txt"),
    extras_require={
        "extras": required("requirements/extras.txt")
    },
    package_data={'': package_files('ovos_utils')},
    include_package_data=True,
    license='Apache',
    author='jarbasAI',
    author_email='jarbas@openvoiceos.com',
    description='collection of simple utilities for use across the openvoiceos ecosystem',
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        'console_scripts': [
            'ovos-logs=ovos_utils.log_parser:ovos_logs'
        ]
    }
)
